from inspect import signature

from .logger import logger
from .markdown import markdown_to_notion, notion_to_markdown

class mapper(property):
    def __init__(self, path, python_to_api, api_to_python, *args, **kwargs):
        self.python_to_api = python_to_api
        self.api_to_python = api_to_python
        self.path = (
            ".".join(map(str, path))
            if isinstance(path, list) or isinstance(path, tuple)
            else path
        )
        super().__init__(*args, **kwargs)


def field_map(path, python_to_api=lambda x: x, api_to_python=lambda x: x):
    """
    Returns a property that maps a Block attribute onto a field in the API data structures.
    
    - `path` can either be a top-level field-name, a list that specifies the key names to traverse,
      or a dot-delimited string representing the same traversal.

    - `python_to_api` is a function that converts values as given in the Python layer into the
      internal representation to be sent along in the API request.

    - `api_to_python` is a function that converts what is received from the API into an internal
      representation to be returned to the Python layer.
    """

    if isinstance(path, str):
        path = path.split(".")

    def fget(self):
        kwargs = {}
        from icecream import ic
        if (
            "client" in signature(api_to_python).parameters
            and "id" in signature(api_to_python).parameters
        ):
            kwargs["client"] = self._client
            kwargs["id"] = self.id
        return api_to_python(self.get(path), **kwargs)

    def fset(self, value):
        from icecream import ic
        kwargs = {}
        if "client" in signature(python_to_api).parameters:
            kwargs["client"] = self._client
        self.set(path, python_to_api(value, **kwargs))

    return mapper(
        fget=fget,
        fset=fset,
        path=path,
        python_to_api=python_to_api,
        api_to_python=api_to_python,
    )


def property_map(
    name, python_to_api=lambda x: x, api_to_python=lambda x: x, markdown=True
):
    """
    Similar to `field_map`, except it works specifically with the data under the "properties" field
    in the API's block table, and just takes a single name to specify which subkey to reference.
    Also, these properties all seem to use a special "embedded list" format that breaks the text
    up into a sequence of chunks and associated format metadata. If `markdown` is True, we convert
    this representation into commonmark-compatible markdown, and back again when saving.
    """

    def py2api(x, client=None):
        kwargs = {}
        if "client" in signature(python_to_api).parameters:
            kwargs["client"] = client
        x = python_to_api(x, **kwargs)
        if markdown:
            x = markdown_to_notion(x)
        return x

    def api2py(x, client=None, id=""):
        x = x or [[""]]
        if markdown:
            x = notion_to_markdown(x)
        kwargs = {}
        params = signature(api_to_python).parameters
        if "client" in params:
            kwargs["client"] = client
        if "id" in params:
            kwargs["id"] = id
        return api_to_python(x, **kwargs)

    return field_map(["properties", name], python_to_api=py2api, api_to_python=api2py)


def joint_map(*mappings):
    """
    Combine multiple `field_map` and `property_map` instances together to map an attribute to multiple API fields.
    Note: when "getting", the first one will be used. When "setting", they will all be set in parallel.
    """

    def fget(self):
        return mappings[0].fget(self)

    def fset(self, value):
        for m in mappings:
            m.fset(self, value)

    return property(fget=fget, fset=fset)



class CriticalError(Exception):
    """Exception raised when a generic critical error occurs."""


class NotionError(Exception):
    """Exception raised when a Notion related critical error occurs."""


class TypeConversionError(Exception):
    """Exception raised when a type conversion error occurs."""



import logging
from argparse import Namespace
from functools import partial
from typing import List
from icecream import ic

from tqdm import tqdm

from csv2notion_neo.local_data import LocalData
from csv2notion_neo.notion_convert import NotionRowConverter
from csv2notion_neo.notion_db import NotionDB, notion_db_from_csv
from csv2notion_neo.notion_db_client import NotionClientExtended
from csv2notion_neo.notion_preparator import NotionPreparator
from csv2notion_neo.notion_uploader import NotionUploadRow
from csv2notion_neo.utils_static import ConversionRules
from csv2notion_neo.utils_threading import ThreadRowUploader, process_iter

logger = logging.getLogger(__name__)


def new_database(
    args: Namespace, client: NotionClientExtended, csv_data: LocalData
) -> str:
    skip_columns = []
    if args.image_column and not args.image_column_keep:
        skip_columns.append(args.image_column)
    if args.icon_column and not args.icon_column_keep:
        skip_columns.append(args.icon_column)
    if args.image_caption_column and not args.image_caption_column_keep:
        skip_columns.append(args.image_caption_column)

    logger.info("Creating new database")

    url, collection_id = notion_db_from_csv(
        client,
        page_name=args.csv_file.stem,
        csv_data=csv_data,
        skip_columns=skip_columns,
    )

    logger.info(f"New database URL: {url}")

    return url


def convert_csv_to_notion_rows(
    csv_data: LocalData, client: NotionClientExtended, collection_id: str, args: Namespace
) -> List[NotionUploadRow]:
    notion_db = NotionDB(client, collection_id)

    conversion_rules = ConversionRules.from_args(args)

    NotionPreparator(notion_db, csv_data, conversion_rules).prepare()


    converter = NotionRowConverter(notion_db, conversion_rules)
    
    return converter.convert_to_notion_rows(csv_data)


def upload_rows(
    notion_rows: List[NotionUploadRow],
    client: NotionClientExtended,
    collection_id: str,
    is_merge: bool,
    max_threads: int,
) -> None:
    worker = partial(
        ThreadRowUploader(client, collection_id).worker,
        is_merge=is_merge,
    )

    tdqm_iter = tqdm(
        iterable=process_iter(worker, notion_rows, max_workers=max_threads),
        total=len(notion_rows),
        leave=False,
    )

    # Consume iterator
    list(tdqm_iter)



import random
import string
from typing import Iterable, List, Set


def rand_id(id_length: int) -> str:
    rand_chars = random.sample(string.digits + string.ascii_letters, id_length)
    return "".join(rand_chars)


def rand_id_list(size: int, id_length: int) -> List[str]:
    """Generate a list of guaranteed unique ids"""

    rand_ids: Set[str] = set()

    while len(rand_ids) < size:
        rand_ids.add(rand_id(id_length))

    return list(rand_ids)


def rand_id_unique(size: int, existing_ids: Iterable[str]) -> str:
    existing_ids = set(existing_ids)

    while True:
        new_id = rand_id(size)
        if new_id not in existing_ids:
            return new_id



from argparse import Namespace
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

ALLOWED_TYPES = frozenset(
    (
        "checkbox",
        "date",
        "multi_select",
        "status",
        "select",
        "number",
        "email",
        "phone_number",
        "url",
        "text",
        "created_time",
        "last_edited_time",
        "created_by",
        "last_edited_by",
        "rollup",
        "formula",
        "file",
        "person",
    )
)

UNSETTABLE_TYPES = frozenset(("created_by", "last_edited_by", "rollup", "formula"))

FileType = Union[str, Path]


@dataclass
class ConversionRules(object):
    csv_file: Path

    image_column: Optional[str]
    image_column_keep: bool
    image_column_mode: str
    image_caption_column: Optional[str]
    image_caption_column_keep: bool

    icon_column: Optional[str]
    icon_column_keep: bool
    default_icon: Optional[FileType]

    merge: bool
    merge_only_column: List[str]
    merge_skip_new: bool

    add_missing_columns: bool
    add_missing_relations: bool

    mandatory_column: List[str]
    fail_on_relation_duplicates: bool
    fail_on_duplicates: bool
    fail_on_conversion_error: bool
    fail_on_inaccessible_relations: bool
    fail_on_missing_columns: bool
    fail_on_unsettable_columns: bool
    fail_on_wrong_status_values: bool

    rename_notion_key_column: List[str]
    payload_key_column: Optional[str]

    hugging_face_token: Optional[str]
    caption_column: List[str]
    hf_model:Optional[str]

    @property
    def files_search_path(self) -> Path:
        return self.csv_file.parent

    @classmethod
    def from_args(cls, args: Namespace) -> "ConversionRules":
        args_map = {
            arg_name: getattr(args, arg_name) for arg_name in cls.__dataclass_fields__
        }

        return cls(**args_map)



import random
from typing import Any, Dict, List, Optional, Tuple, cast

from csv2notion_neo.notion.collection import Collection, NotionSelect, CollectionRowBlock, CalendarView
from csv2notion_neo.notion.operations import build_operation

from csv2notion_neo.notion_row import CollectionRowBlockExtended
from csv2notion_neo.utils_db import make_status_column
from csv2notion_neo.utils_rand_id import rand_id_unique
from icecream import ic

class CollectionExtended(Collection):
    def get_rows(self) -> List[CollectionRowBlockExtended]:  # noqa: WPS615
        return [
            CollectionRowBlockExtended(row._client, row._id)
            for row in super().get_rows()
        ]

    def get_unique_rows(self) -> Dict[str, CollectionRowBlockExtended]:
        rows: Dict[str, CollectionRowBlockExtended] = {}

        # sort rows so that only first row is kept if multiple have same title
        sorted_rows = sorted(self.get_rows(), key=lambda r: str(r.title))

        for row in sorted_rows:
            rows.setdefault(row.title, row)
        
        return rows

    def add_row_block(
        self,
        update_views: bool = True,
        row_class: Optional[type] = None,
        properties: Optional[Dict[str, Any]] = None,
        columns: Optional[Dict[str, Any]] = None,
    ) -> CollectionRowBlockExtended:
        row_class = row_class or CollectionRowBlockExtended

        new_row = self._add_row_block(
            update_views=update_views,
            row_class=row_class,
            properties=properties,
            columns=columns,
        )

        return cast(CollectionRowBlockExtended, new_row)

    def _add_row_block(
        self,
        update_views: bool = True,
        row_class: Optional[type] = None,
        properties: Optional[Dict[str, Any]] = None,
        columns: Optional[Dict[str, Any]] = None,
    ) -> CollectionRowBlock:
        """
        Create a new empty CollectionRowBlock under this collection, and return the instance.
        """

        row_class = row_class or CollectionRowBlock

        row_id = self._client.create_record("block", self, type="page")
        row = row_class(self._client, row_id)

        columns = {} if columns is None else columns
        properties = {} if properties is None else properties

        with self._client.as_atomic_transaction():
            for key, val in properties.items():
                setattr(row, key, val)

        
            for key, val in columns.items():
                setattr(row.columns, key, val)

            if update_views:
                # make sure the new record is inserted at the end of each view
                for view in self.parent.views:
                    if view is None or isinstance(view, CalendarView):
                        continue

                    page_sort = view.get("page_sort", [])

                    if not page_sort:
                        view.set("page_sort", [row_id])
                    else:
                        last_element = page_sort[-1]
                        self._client.submit_transaction(
                            build_operation(
                                id=view.id,
                                path=["page_sort"],
                                args={"id": row_id, "after": last_element},
                                command="listAfter",
                                table=view._table,
                            )
                        )

        return row

    def add_column(self, column_name: str, column_type: str) -> None:
        schema_raw = self.get("schema")
        new_id = rand_id_unique(4, schema_raw)
        schema_raw[new_id] = {"name": column_name, "type": column_type}

        if column_type == "status":
            schema_raw[new_id].update(make_status_column())

        self.set("schema", schema_raw)

    def has_duplicates(self) -> bool:
        row_titles = [row.title for row in self.get_rows()]
        return len(row_titles) != len(set(row_titles))

    def is_accessible(self) -> bool:
        rec = self._client.get_record_data("collection", self.id, force_refresh=True)
        return rec is not None

    def check_schema_select_options(  # noqa: WPS210
        self, prop: Dict[str, Any], values: Any  # noqa: WPS110
    ) -> Tuple[bool, Dict[str, Any]]:
        schema_update = False

        prop_options = prop.setdefault("options", [])
        current_options = [p["value"].lower() for p in prop_options]
        if not isinstance(values, list):
            values = [values]  # noqa: WPS110

        for v in values:
            if v and v.lower() not in current_options:
                schema_update = True

                if self._client.options.get("is_randomize_select_colors") is True:
                    color = _get_random_select_color()
                else:
                    color = "default"

                prop_options.append(NotionSelect(v, color).to_dict())
        return schema_update, prop


def _get_random_select_color() -> str:
    return str(random.choice(NotionSelect.valid_colors))  # noqa: S311



import argparse
from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple, Union

from csv2notion_neo.notion_convert_map import map_icon
from csv2notion_neo.utils_exceptions import CriticalError
from csv2notion_neo.utils_static import ALLOWED_TYPES, FileType
from csv2notion_neo.utils_str import split_str
from csv2notion_neo.version import __version__

ArgToken = Union[str, Tuple[str, str]]
ArgOption = Dict[str, Any]
ArgSchema = Dict[str, Dict[ArgToken, ArgOption]]
HELP_ARGS_WIDTH = 50

class CustomHelpFormatter(argparse.RawTextHelpFormatter):
    def _format_args(self, action, default_metavar):
        return ""
    
def parse_args(argv: Sequence[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="csv2notion_neo",
        description="https://github.com/TheAcharya/csv2notion-neo \n\nUpload & Merge CSV or JSON Data with Images to Notion Database",
        usage="%(prog)s [-h] --token TOKEN [--url URL] [OPTION]... FILE",
        add_help=False,
        formatter_class=lambda prog: CustomHelpFormatter(
            prog, max_help_position=HELP_ARGS_WIDTH
        ),
    )

    schema: ArgSchema = {
        "POSITIONAL": {
            "csv_file": {
                "type": Path,
                "help": "CSV or JSON file to upload",
                "metavar": "FILE",
            }
        },
        "general options": {
            "--workspace":{
                "help": (
                    "active Notion workspace name"
                ),
                "required":True,
                "metavar": "workspace",
            },
            "--token": {
                "help": "Notion token, stored in token_v2 cookie for notion.so",
                "required": True,
            },
            "--url": {
                "help": (
                    "Notion database URL; if none is provided,"
                    " will create a new database"
                ),
                "metavar": "URL",
            },
            "--max-threads": {
                "type": lambda x: max(int(x), 1),
                "default": 5,
                "help": "upload threads (default: 5)",
                "metavar": "NUMBER",
            },
            "--log": {
                "type": Path,
                "metavar": "FILE",
                "help": "file to store program log",
            },
            "--verbose": {
                "action": "store_true",
                "help": "output debug information",
            },
            "--version": {
                "action": "version",
                "version": f"%(prog)s {__version__}",
            },
            ("-h", "--help"): {
                "action": "help",
                "help": "show this help message and exit",
            },
        },
        "machine learning options":{
            "--hugging-face-token":{
                "help": (
                    "Hugging Face token to use image captioning model online"
                ),
                "metavar": "AI",
            },
            "--hf-model":{
                "help": (
                    "Provide the model used for generating caption <vit-gpt2 | blip-image | git-large> (defaults: vit-gpt2)"
                ),
                "metavar": "AI",
            },
            "--caption-column":{
                "help":(
                    "Provide both image column and column where caption would be written"
                ),
                "metavar":"AI",
                "nargs":2,
            }
        },
        "column options": {
            "--column-types": {
                "help": (
                    "comma-separated list of column types to use for non-key columns;"
                    "\nif none is provided, types will be guessed from CSV values"
                    "\n(can also be used with --add-missing-columns flag)"
                ),
                "metavar": "TYPES",
                "type": _parse_column_types,
            },
            "--delimiter": {
                "help": (
                    "Delimiter that is used to seperate columns and rows in csv file"
                    "\nif none is provided, ',' is taken as default"
                ),
                "default": ","
            },
            "--add-missing-columns": {
                "action": "store_true",
                "help": (
                    "if columns are present in CSV or JSON but not in Notion DB,"
                    " add them to Notion DB"
                ),
            },
            "--rename-notion-key-column":{
                "nargs":2,
                "help":"rename the key column in the file to a different key column in Notion DB",
                "metavar":"column",
            },
            "--randomize-select-colors": {
                "action": "store_true",
                "help": (
                    "randomize colors for added options"
                    " in select and multi select columns"
                ),
            },
        },
        "merge options": {
            "--merge": {
                "action": "store_true",
                "help": (
                    "merge CSV or JSON with existing Notion DB rows,"
                    " first column will be used as a key"
                ),
            },
            "--merge-only-column": {
                "action": "append",
                "help": (
                    "CSV or JSON column that should be updated on merge;"
                    "\nwhen provided, other columns will be ignored"
                    "\n(use multiple times for multiple columns)"
                ),
                "metavar": "COLUMN",
                "default": [],
            },
            "--merge-skip-new": {
                "action": "store_true",
                "help": (
                    "skip new rows in CSV or JSON that are not already in Notion DB"
                    " during merge"
                ),
            },
        },
        "relations options": {
            "--add-missing-relations": {
                "action": "store_true",
                "help": "add missing entries into linked Notion DB",
            },
        },
        "page cover options": {
            "--image-column": {
                "help": (
                    "One or more CSV or JSON column that points to URL or image file"
                    " that will be embedded for that row"
                ),
                "nargs":"*",
                "metavar": "COLUMN",
            },
            "--image-column-keep": {
                "action": "store_true",
                "help": "keep image CSV or JSON column as a Notion DB column",
            },
            "--image-column-mode": {
                "choices": ["cover", "block"],
                "default": "block",
                "help": (
                    "upload image as [cover] or insert it as [block]"
                    " (default: block)"
                ),
            },
            "--image-caption-column": {
                "help": (
                    "CSV or JSON column that points to text caption"
                    " that will be added to the image block"
                    "\nif --image-column-mode is set to 'block'"
                ),
                "metavar": "COLUMN",
            },
            "--image-caption-column-keep": {
                "action": "store_true",
                "help": "keep image caption CSV column as a Notion DB column",
            },
        },
        "page icon options": {
            "--icon-column": {
                "help": (
                    "CSV or JSON column that points to emoji, URL or image file"
                    "\nthat will be used as page icon for that row"
                ),
                "metavar": "COLUMN",
            },
            "--icon-column-keep": {
                "action": "store_true",
                "help": "keep icon CSV or JSON column as a Notion DB column",
            },
            "--default-icon": {
                "help": (
                    "Emoji, URL or image file"
                    " that will be used as page icon for every row by default"
                ),
                "metavar": "ICON",
                "type": _parse_default_icon,
            },
        },
        "validation options": {
            "--mandatory-column": {
                "action": "append",
                "help": (
                    "CSV or JSON column that cannot be empty"
                    " (use multiple times for multiple columns)"
                ),
                "metavar": "COLUMN",
                "default": [],
            },
            "--payload-key-column": {
                "help": (
                    "JSON object that is the key in notion db."
                    " if json file is used, this cannot be empty!"
                ),
                "metavar": "key column"
            },
            "--fail-on-relation-duplicates": {
                "action": "store_true",
                "help": (
                    "fail if any linked DBs in relation columns have duplicate entries;"
                    "\notherwise, first entry in alphabetical order"
                    "\nwill be treated as unique when looking up relations"
                ),
            },
            "--fail-on-duplicates": {
                "action": "store_true",
                "help": (
                    "fail if Notion DB or CSV has duplicates in key column,"
                    "\nuseful when sanitizing before merge to avoid ambiguous mapping"
                ),
            },
            "--fail-on-duplicate-csv-columns": {
                "action": "store_true",
                "help": (
                    "fail if CSV or JSON has duplicate columns;"
                    "\notherwise last column will be used"
                ),
            },
            "--fail-on-conversion-error": {
                "action": "store_true",
                "help": (
                    "fail if any column type conversion error occurs;"
                    "\notherwise errors will be replaced with empty strings"
                ),
            },
            "--fail-on-inaccessible-relations": {
                "action": "store_true",
                "help": (
                    "fail if any relation column points to a Notion DB that"
                    "\nis not accessible to the current user;"
                    "\notherwise those columns will be ignored"
                ),
            },
            "--fail-on-missing-columns": {
                "action": "store_true",
                "help": (
                    "fail if columns are present in CSV or JSON but not in Notion DB;"
                    "\notherwise those columns will be ignored"
                ),
            },
            "--fail-on-unsettable-columns": {
                "action": "store_true",
                "help": (
                    "fail if DB has columns that don't support assigning value to them;"
                    "\notherwise those columns will be ignored"
                    "\n(columns with type created_by, last_edited_by,"
                    " rollup or formula)"
                ),
            },
            "--fail-on-wrong-status-values": {
                "action": "store_true",
                "help": (
                    "fail if values for 'status' columns don't have"
                    " matching option in DB;"
                    "\notherwise those values will be replaced with default status"
                ),
            },
        },
    }

    _parse_schema(parser, schema)

    return parser.parse_args(argv)


def _parse_schema(  # noqa: WPS210
    parser: argparse.ArgumentParser, schema: ArgSchema
) -> None:
    group: argparse._ActionsContainer

    for group_name, group_args in schema.items():
        if group_name == "POSITIONAL":
            group = parser
        else: 
            group = parser.add_argument_group(group_name)

        for arg, arg_params in group_args.items():
            opt_arg = [arg] if isinstance(arg, str) else arg
            group.add_argument(*opt_arg, **arg_params)


def _parse_default_icon(default_icon: str) -> FileType:
    default_icon_filetype = map_icon(default_icon)
    if isinstance(default_icon_filetype, Path):
        if not default_icon_filetype.exists():
            raise CriticalError(f"File not found: {default_icon_filetype}")
    return default_icon_filetype


def _parse_column_types(column_types: str) -> List[str]:

    column_types_list = split_str(column_types)
    unknown_types = set(column_types_list) - set(ALLOWED_TYPES)
    if unknown_types:
        raise CriticalError(
            "Unknown types: {0}; allowed types: {1}".format(
                ", ".join(unknown_types), ", ".join(ALLOWED_TYPES)
            )
        )
    return column_types_list



from dataclasses import dataclass
from typing import Any, Dict

from csv2notion_neo.notion_db import NotionDB
from csv2notion_neo.notion_row import CollectionRowBlockExtended
from icecream import ic
from csv2notion_neo.utils_ai import AI

@dataclass
class NotionUploadRow(object):
    columns: Dict[str, Any]
    properties: Dict[str, Any]

    def key(self) -> str:
        if "payload_key_column" in self.properties:
            if self.properties["payload_key_column"]:
                return str(self.columns[self.properties["payload_key_column"]])
        return str(list(self.columns.values())[0])

class NotionRowUploader(object):
    def __init__(self, db: NotionDB):
        self.db = db

    def upload_row(self, row: NotionUploadRow, is_merge: bool) -> None:

        if 'AI' in row.properties:
            ai_client = AI(row.properties['AI'])
            row.columns = ai_client.out(row.columns)
        #row.columns['ai caption'] = '2'
        post_properties = _extract_post_properties(row.properties)

        db_row = self._get_db_row(row, is_merge)

        # these need to be updated after
        # because they can't be updated in atomic transaction
        for prop, prop_val in post_properties.items():
            setattr(db_row, prop, prop_val)

    def _get_db_row(
        self, row: NotionUploadRow, is_merge: bool
    ) -> CollectionRowBlockExtended:
      
        existing_row = self.db.rows.get(row.key()) if is_merge else None

        if is_merge and existing_row:
            cur_row = existing_row
            cur_row.update(properties=row.properties, columns=row.columns)
        else:
            cur_row = self.db.add_row(properties=row.properties, columns=row.columns)
        return cur_row


def _extract_post_properties(properties: Dict[str, Any]) -> Dict[str, Any]:
    return {
        p: properties.pop(p)
        for p in properties.copy()
        if p in {"cover_block", "cover_block_caption", "last_edited_time","cover"}
    }



import threading
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Callable, Iterable, Iterator

from csv2notion_neo.notion_db import NotionDB
from csv2notion_neo.notion_db_client import NotionClientExtended
from csv2notion_neo.notion_uploader import NotionRowUploader
from icecream import ic

class ThreadRowUploader(object):
    def __init__(self, client: NotionClientExtended, collection_id: str) -> None:
        self.thread_data = threading.local()

        self.client = client
        self.collection_id = collection_id

    def worker(self, *args: Any, **kwargs: Any) -> None:
        try:
            notion_uploader = self.thread_data.uploader
        except AttributeError:
            client = NotionClientExtended(old_client=self.client)
            notion_db = NotionDB(client, self.collection_id)
            
            notion_uploader = NotionRowUploader(notion_db)
            self.thread_data.uploader = notion_uploader

        notion_uploader.upload_row(*args, **kwargs)


def process_iter(
    worker: Callable[[Any], None], tasks: Iterable[Any], max_workers: int
) -> Iterator[None]:
    if max_workers == 1:
        yield from map(worker, tasks)
    else:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(worker, t) for t in tasks]

            yield from (f.result() for f in as_completed(futures))



from typing import Any, Optional, cast

from csv2notion_neo.notion.block import ImageBlock
from csv2notion_neo.notion.collection import CollectionRowBlock
from csv2notion_neo.notion.maps import field_map
from csv2notion_neo.notion.operations import build_operation

from csv2notion_neo.notion_row_upload_file import get_file_id
from icecream import ic


class CoverImageBlock(ImageBlock):
    is_cover_block = field_map("properties.is_cover_block")

    
    def update(self, **attrs: Any) -> None:
        with self._client.as_atomic_transaction():
            file_id = attrs.pop("file_id", None)

            for k, v in attrs.items():
                setattr(self, k, v)

            if file_id:
                self._client.submit_transaction(
                    build_operation(
                        id=self.id,
                        path=["file_ids"],
                        args={"id": file_id},
                        command="listAfter",
                    )
                )


class RowCoverImageBlock(object):  # noqa: WPS214
    def __init__(self, row: CollectionRowBlock):
        self.row = row

        self._image_block: Optional[CoverImageBlock] = None

    @property
    def image_block(self) -> Optional[CoverImageBlock]:
        if self._image_block is None:
            self._image_block = self._get_cover_image_block()

        return self._image_block

    @image_block.setter
    def image_block(self, image_block: CoverImageBlock) -> None:
        self._image_block = image_block

    @property
    def caption(self) -> Optional[str]:
        if self.image_block is None:
            return None

        caption = self.image_block.caption

        return str(caption) if caption is not None else None

    @caption.setter
    def caption(self, caption: Optional[str]) -> None:
        if self.image_block is None:
            return

        self.image_block.caption = caption

    @property
    def url(self) -> str:
        if self.image_block is None:
            return ""

        return str(self.image_block.source)

    @url.setter
    def url(self, image_url: Optional[str]) -> None:
        
        cover_img = image_url.pop(0)
        if cover_img is None:
            if self.image_block is not None:
                self.image_block.remove()
            return

        attrs = {
            "display_source": cover_img,
            "source": cover_img,
        }


        file_id = get_file_id(cover_img)
        if file_id:
            attrs["file_id"] = file_id

        if self.row.children:
            if self.image_block:
                self.image_block.update(**attrs)
            else:
                self.image_block = self._add_new_image_block(**attrs)
                self.image_block.move_to(self.row, "first-child")
        else:
            self.image_block = self._add_new_image_block(**attrs)


        self.image_block.is_cover_block = True

        if image_url:
            for img in image_url:
                if img is None:
                    if self.image_block is not None:
                        self.image_block.remove()
                    return
                attrs = {
                    "display_source": img,
                    "source": img,
                }


                file_id = get_file_id(img)
                if file_id:
                    attrs["file_id"] = file_id

                self.new_image_block = self._add_new_image_block(**attrs)

    def _add_new_image_block(self, **attrs: Any) -> CoverImageBlock:
      
        image_block = self.row.children.add_new(ImageBlock, **attrs)
        # image_block = CoverImageBlock(image_block._client, image_block._id)
        return cast(CoverImageBlock, image_block)

    def _get_cover_image_block(self) -> Optional[CoverImageBlock]:
        if not self.row.children:
            return None

        image_block = self.row.children[0]
        if not isinstance(image_block, ImageBlock):
            return None

        image_block = CoverImageBlock(image_block._client, image_block._id)

        if not image_block.is_cover_block:
            return None

        return cast(CoverImageBlock, image_block)



from .utils import now
from icecream import ic


def build_operation(id, path, args, command="set", table="block"):
    """
    Data updates sent to the submitTransaction endpoint consist of a sequence of "operations". This is a helper
    function that constructs one of these operations.
    """

    if isinstance(path, str):
        path = path.split(".")

    return {"id": id, "path": path, "args": args, "command": command, "table": table}


def operation_update_last_edited(user_id, block_id):
    """
    When transactions are submitted from the web UI, it also includes an operation to update the "last edited"
    fields, so we want to send those too, for consistency -- this convenience function constructs the operation.
    """
    return {
        "args": {
            "last_edited_by_id": user_id,
            "last_edited_by_table": "notion_user",
            "last_edited_time": now(),
        },
        "command": "update",
        "id": block_id,
        "path": [],
        "table": "block",
    }



from copy import deepcopy
from typing import Any, Dict, Optional

from csv2notion_neo.notion.client import NotionClient, create_session
from csv2notion_neo.notion.space import Space
from csv2notion_neo.notion.store import RecordStore
from csv2notion_neo.notion.user import User

from csv2notion_neo.notion_db_collection import CollectionExtended
from icecream import ic

class NotionClientExtended(NotionClient):
    def __init__(
        self,
        *args: Any,
        old_client: Optional["NotionClientExtended"] = None,
        options: Optional[Dict[str, Any]] = None,
        workspace = None,
        **kwargs: Any,
    ):
        
        self.options = options or {}

        if old_client is None:
            super().__init__(*args, **kwargs,workspace=workspace)
            return

        self._monitor = None


        self.session = create_session()
        self.session.cookies = old_client.session.cookies.copy()

        self._store = self._clone_store(old_client)

        self._clone_user_info(old_client)

        self.options = old_client.options.copy()

    def get_collection(
        self, collection_id: str, force_refresh: bool = False
    ) -> Optional[CollectionExtended]:
        coll = self.get_record_data(
            "collection", collection_id, force_refresh=force_refresh
        )
        return CollectionExtended(self, collection_id) if coll else None

    def _clone_store(self, old_client: NotionClient) -> RecordStore:
        new_store = RecordStore(self)
        old_store = old_client._store

        new_store._values = deepcopy(old_store._values)
        new_store._role = deepcopy(old_store._role)
        new_store._collection_row_ids = deepcopy(old_store._collection_row_ids)

        return new_store

    def _clone_user_info(self, old_client: NotionClient) -> None:
        self.current_user = User(self, old_client.current_user.id)
        self.current_space = Space(self, old_client.current_space.id)

        self.session.headers.update(
            {"x-notion-active-user-header": self.current_user.id}
        )



import ssl
import sys

from requests.adapters import HTTPAdapter
from urllib3.util.ssl_ import create_urllib3_context

# https://github.com/encode/httpx/blob/0.23.0/httpx/_compat.py
if sys.version_info >= (3, 10) or (
    sys.version_info >= (3, 7) and ssl.OPENSSL_VERSION_INFO >= (1, 1, 0, 7)
):

    def set_minimum_tls_version_1_2(context: ssl.SSLContext) -> None:
        # The OP_NO_SSL* and OP_NO_TLS* become deprecated in favor of
        # 'SSLContext.minimum_version' from Python 3.7 onwards, however
        # this attribute is not available unless the ssl module is compiled
        # with OpenSSL 1.1.0g or newer.
        # https://docs.python.org/3.10/library/ssl.html#ssl.SSLContext.minimum_version
        # https://docs.python.org/3.7/library/ssl.html#ssl.SSLContext.minimum_version
        context.minimum_version = ssl.TLSVersion.TLSv1_2

else:

    def set_minimum_tls_version_1_2(context: ssl.SSLContext) -> None:
        # If 'minimum_version' isn't available, we configure these options with
        # the older deprecated variants.
        context.options |= ssl.OP_NO_SSLv2
        context.options |= ssl.OP_NO_SSLv3
        context.options |= ssl.OP_NO_TLSv1
        context.options |= ssl.OP_NO_TLSv1_1


def create_custom_urllib3_context():
    """
    needed to avoid cloudflare bot detector that returns 403 on some requests
    this makes the requests library have the same TLS fingerprint as httpx
    in the future cloudflare may ban this signature also
    in that case we can use different cipher suite to alter TLS signature even further
    https://stackoverflow.com/questions/64967706/python-requests-https-code-403-without-but-code-200-when-using-burpsuite
    """

    CIPHERS = ('DEFAULT:@SECLEVEL=2')

    context = create_urllib3_context(ciphers=CIPHERS)
    set_minimum_tls_version_1_2(context)
    context.options &= ~ssl.OP_NO_TICKET

    return context


class HTTPAdapterTLS(HTTPAdapter):
    def init_poolmanager(self, *args, **kwargs):
        kwargs["ssl_context"] = create_custom_urllib3_context()
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args, **kwargs):
        kwargs["ssl_context"] = create_custom_urllib3_context()
        return super().proxy_manager_for(*args, **kwargs)



from typing import Any, Dict, List, Optional, Tuple

import requests
from csv2notion_neo.notion.user import User
from csv2notion_neo.notion.utils import InvalidNotionIdentifier

from csv2notion_neo.local_data import LocalData
from csv2notion_neo.notion_db_client import NotionClientExtended
from csv2notion_neo.notion_db_collection import CollectionExtended
from csv2notion_neo.notion_row import CollectionRowBlockExtended
from csv2notion_neo.utils_db import make_status_column
from csv2notion_neo.utils_exceptions import NotionError
from csv2notion_neo.utils_rand_id import rand_id_list
from icecream import ic


class NotionDB(object):  # noqa: WPS214
    def __init__(self, client: NotionClientExtended, collection_id: str):
        self.client = client
        self.collection = CollectionExtended(self.client, collection_id)

        self._cache_columns: Dict[str, Dict[str, str]] = {}
        self._cache_relations: Dict[str, NotionDB] = {}
        self._cache_rows: Dict[str, CollectionRowBlockExtended] = {}
        self._cache_users: Dict[str, User] = {}

    @property
    def name(self) -> str:
        return str(self.collection.name)

    @property
    def columns(self) -> Dict[str, Dict[str, str]]:
        if not self._cache_columns:
            self._cache_columns = {
                p["name"]: p for p in self.collection.get_schema_properties()
            }

        return self._cache_columns

    @property
    def key_column(self) -> str:
        column_values = self.columns.values()
        return next(c["name"] for c in column_values if c["type"] == "title")

    @property
    def rows(self) -> Dict[str, CollectionRowBlockExtended]:
        if not self._cache_rows:
            self._cache_rows = self.collection.get_unique_rows()


        return self._cache_rows

    @property
    def relations(self) -> Dict[str, "NotionDB"]:
        if not self._cache_relations:
            relations = [c for c in self.columns.values() if c["type"] == "relation"]

            self._cache_relations = {
                r["name"]: NotionDB(self.client, r["collection_id"]) for r in relations
            }

        return self._cache_relations

    @property
    def users(self) -> Dict[str, User]:
        if not self._cache_users:
            self._cache_users = {u.email: u for u in self.client.current_space.users}

        return self._cache_users

    def get_user_by_name(self, name: str) -> Optional[User]:
        name_match = (u for u in self.users.values() if u.name == name)
        return next(name_match, None)

    def find_user(self, email: str) -> Optional[User]:
        res = self.client.post("findUser", {"email": email}).json()

        try:
            user_id = res["value"]["value"]["id"]
        except KeyError:
            return None

        found_user = User(self.client, user_id)

        self.users[found_user.email] = found_user

        return found_user

    def has_duplicates(self) -> bool:
        return self.collection.has_duplicates()

    def is_accessible(self) -> bool:
        return self.collection.is_accessible()

    def add_column(self, column_name: str, column_type: str) -> None:
        self.collection.add_column(column_name, column_type)

        self._cache_columns = {}
        self._cache_rows = {}

    def add_row(
        self,
        properties: Optional[Dict[str, Any]] = None,
        columns: Optional[Dict[str, Any]] = None,
    ) -> CollectionRowBlockExtended:
        new_row = self.collection.add_row_block(properties=properties, columns=columns)
        key = columns.get(self.key_column) if columns else None
        if key:
            self.rows[key] = new_row

        return new_row

    def add_row_key(self, key: str) -> CollectionRowBlockExtended:
        return self.add_row(columns={self.key_column: key})


def get_collection_id(client: NotionClientExtended, notion_url: str) -> str:
    try:
        block = client.get_block(notion_url, force_refresh=True)
    except InvalidNotionIdentifier as e:
        raise NotionError("Invalid URL.") from e

    if not block:
        raise NotionError(f"Cannot access {notion_url}.")

    if block.type not in ["collection_view_page", "collection_view"]:
        raise NotionError("Provided URL does not point to a Notion database.")

    if block.collection.role not in {"editor", "content_only_editor", "read_and_write"}:
        raise NotionError("You must have editing permissions for Notion database.")

    return str(block.collection.id)


def notion_db_from_csv(
    client: NotionClientExtended,
    page_name: str,
    csv_data: LocalData,
    skip_columns: Optional[List[str]] = None,
) -> Tuple[str, str]:
    schema = _schema_from_csv(csv_data, skip_columns)

    page_id = client.create_record(
        "block",
        client.current_space,
        type="collection_view_page",
        permissions=[
            {
                "role": "editor",
                "type": "user_permission",
                "user_id": client.current_user.id,
            }
        ],
    )
    page = client.get_block(page_id)

    page.collection = client.get_collection(
        client.create_record(
            "collection",
            parent=page,
            schema=schema,
        )
    )

    if page.collection is None:  # pragma: no cover
        raise NotionError("Failed to create collection.")

    view = page.views.add_new(view_type="table")

    # Make sure all columns are in the same order as CSV
    table_properties = [{"visible": True, "property": col_id} for col_id in schema]
    view.set("format.table_properties", table_properties)

    page.title = page_name

    return str(page.get_browseable_url()), page.collection.id


def _schema_from_csv(
    csv_data: LocalData, skip_columns: Optional[List[str]] = None
) -> Dict[str, Dict[str, str]]:
    if skip_columns:
        columns = [c for c in csv_data.content_columns if c not in skip_columns]
    else:
        columns = csv_data.content_columns

    schema_ids = rand_id_list(len(columns), 4)

    schema = {"title": {"name": csv_data.key_column, "type": "title"}}

    for col_id, col_key in zip(schema_ids, columns):
        schema[col_id] = {
            "name": col_key,
            "type": csv_data.col_type(col_key),
        }

        if schema[col_id]["type"] == "status":
            schema[col_id].update(make_status_column())

    return schema


def get_notion_client(token: str,workspace: str, **options: Dict[str, Any]) -> NotionClientExtended:
    try:
        client = NotionClientExtended(token_v2=token,workspace=workspace)
    except requests.exceptions.HTTPError as e:
        raise NotionError("Invalid Notion token") from e

    client.options = options

    return client



from typing import List
from icecream import ic

def split_str(s: str, sep: str = ",") -> List[str]:
    
    if type(s) == list:
        return ["".join(item.split(",")) for item in s]
    
    return [v.strip() for v in s.split(sep) if v.strip()]



import logging
import os
import signal
import sys
from pathlib import Path
from typing import Any, Optional
from icecream import ic

from csv2notion_neo.version import __version__
from csv2notion_neo.cli_args import parse_args
from csv2notion_neo.cli_steps import convert_csv_to_notion_rows, new_database, upload_rows
from csv2notion_neo.local_data import LocalData
from csv2notion_neo.notion_db import get_collection_id, get_notion_client
from csv2notion_neo.utils_exceptions import CriticalError, NotionError

logger = logging.getLogger(__name__)


def cli(*argv: str) -> None:
    try:
        ic.enable()
        args = parse_args(argv)
        
        setup_logging(is_verbose=args.verbose, log_file=args.log)
        logger.info(f"CSV2Notion Neo version {__version__}")

        path = Path(args.csv_file).suffix

        if "json" in path:
            if not args.payload_key_column:
                raise CriticalError("Json file found, please enter the key column!")
            
        logger.info(f"Validating {path[1::]} & csv2notion_neo.notion DB schema")

        csv_data = LocalData(
            args.csv_file, args.column_types, args.fail_on_duplicate_csv_columns, args.payload_key_column,args=args
        )

        if not csv_data:
            raise CriticalError(f"{path} file is empty")

        client = get_notion_client(
            args.token,
            workspace=args.workspace,
            is_randomize_select_colors=args.randomize_select_colors,
        )

        if not args.url:
            args.url = new_database(args, client, csv_data)
        
        collection_id = get_collection_id(client, args.url)

        notion_rows = convert_csv_to_notion_rows(csv_data, client, collection_id, args)

        logger.info("Uploading {0}...".format(args.csv_file.name))

        upload_rows(
            notion_rows,
            client=client,
            collection_id=collection_id,
            is_merge=args.merge,
            max_threads=args.max_threads,
        )

        logger.info("Done!")

    except Exception as e:
        if args.verbose:
            logger.error('Error at %s', 'division', exc_info=e)
        else:
            logger.error(e)


def setup_logging(is_verbose: bool = False, log_file: Optional[Path] = None) -> None:
    logging.basicConfig(format="%(levelname)s: %(message)s")

    logging.getLogger("csv2notion_neo").setLevel(
        logging.DEBUG if is_verbose else logging.INFO
    )

    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(
            logging.Formatter("%(asctime)s [%(levelname)-8.8s] %(message)s")
        )
        logging.getLogger("csv2notion_neo").addHandler(file_handler)

    logging.getLogger("csv2notion_neo.notion").setLevel(logging.WARNING)


def abort(*_: Any) -> None:  # pragma: no cover
    print("\nAbort")  # noqa: WPS421
    os._exit(1)


def main() -> None:
    signal.signal(signal.SIGINT, abort)

    try:
        cli(*sys.argv[1:])
    except (NotionError, CriticalError) as e:
        logger.critical(str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(1)



import csv
import json
import logging
from collections import Counter
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional

from csv2notion_neo.notion_type_guess import guess_type_by_values
from csv2notion_neo.utils_exceptions import CriticalError
from icecream import ic

CSVRowType = Dict[str, str]

logger = logging.getLogger(__name__)




def data_read(file_path: Path, fail_on_duplicate_columns: bool,args:Dict) -> List[CSVRowType]:
    suffix = Path(file_path).suffix

    if "csv" in suffix:
        try:
            with open(file_path, "r", encoding="utf-8-sig") as csv_file:
                return _csv_read_rows(csv_file, fail_on_duplicate_columns,args.delimiter)
        except FileNotFoundError as e:
            raise CriticalError(f"File {file_path} not found") from e
    elif "json" in suffix:
        try:
            with open(file_path, "r", encoding="utf-8-sig") as json_file:
                return _json_read_rows(json_file, fail_on_duplicate_columns)
        except FileNotFoundError as e:
            raise CriticalError(f"File {file_path} not found") from e

    else:
        raise CriticalError(f"{suffix} file extension not supported!")


def _csv_read_rows(
    csv_file: Iterable[str], fail_on_duplicate_columns: bool, delemiter: str = ','
) -> List[CSVRowType]:
    
    reader = csv.DictReader(csv_file,delimiter=delemiter,restval="")
    
    if not reader.fieldnames:
        raise CriticalError("CSV file has no columns.")

    duplicate_columns = _list_duplicates(list(reader.fieldnames))

    if duplicate_columns:
        message = f"Duplicate columns found in CSV: {duplicate_columns}."

        if fail_on_duplicate_columns:
            raise CriticalError(message)

        logger.warning(message)

    rows = list(reader)

    if rows and None in rows[0]:
        logger.warning(
            "Inconsistent number of columns detected."
            " Excess columns will be truncated."
        )
        rows = [_drop_dict_columns(row, [None]) for row in rows]

    return rows

def _json_read_rows(
    json_file: Iterable[str], fail_on_duplicate_columns: bool
) -> List[CSVRowType]:
    reader = json.load(json_file)

    if not reader:
        raise CriticalError("JSON file has no data")

    rows = list(reader)
    rows = [_convert_int_to_string(row) for row in rows]

    if rows and None in rows[0]:
        logger.warning(
            "Inconsistent number of columns detected."
            " Excess columns will be truncated."
        )
        rows = [_drop_dict_columns(row, [None]) for row in rows]

    return rows

def _list_duplicates(lst: List[str]) -> List[str]:
    return [lst_item for lst_item, count in Counter(lst).items() if count > 1]


def _drop_dict_columns(
    src_dict: Dict[Any, Any], columns_to_drop: Iterable[Any]
) -> Dict[Any, Any]:
    return {k: v for k, v in src_dict.items() if k not in columns_to_drop}

def _convert_int_to_string(src_dict:Dict[Any,Any]):
    for key,val in src_dict.items():
        if type(val) == int:
            src_dict[key] = str(val)
    
    return src_dict

class LocalData(Iterable[CSVRowType]):  # noqa:  WPS214
    def __init__(
        self,
        csv_file: Path,
        column_types: Optional[List[str]] = None,
        fail_on_duplicate_columns: bool = False,
        key_col_json: str = None,
        args:dict = None
    ) -> None:
        self.csv_file = csv_file
        self.rows = data_read(self.csv_file, fail_on_duplicate_columns, args)
        self.key_col = key_col_json
        self.types = self._column_types(column_types)


        #ai features
        try:
            if args.hugging_face_token:
                if args.caption_column:
                    self._create_ai_columns(args.caption_column[1])
                    self.model_url = self._pick_model(args.hf_model)
                else:
                    raise Exception("ai token provided, please provide column maps to fill the ai content")

        except Exception as e:
            logger.error(f"Problem in creating AI column {e}")

    def __len__(self) -> int:
        return len(self.rows)

    def __iter__(self) -> Iterator[CSVRowType]:
        yield from self.rows

    @property
    def key_column(self) -> str:
        if self.key_col:
            return self.key_col
        
        return self.columns[0]

    @property
    def content_columns(self) -> List[str]:
        if self.key_column:
            return [ele for ele in self.columns if ele != self.key_column]
        return self.columns[1:]

    @property
    def columns(self) -> List[str]:
        return list(self.rows[0].keys()) if self.rows else []

    def columns_of_type(self, col_type: str) -> List[str]:
        return [col for col in self.content_columns if self.col_type(col) == col_type]

    def col_type(self, col_name: str) -> str:
        return self.types[col_name]

    def col_values(self, col_name: str) -> List[str]:
        return [row[col_name] for row in self.rows]

    def drop_columns(self, *columns: str) -> None:
        self.rows = [_drop_dict_columns(row, columns) for row in self.rows]
        self.types = _drop_dict_columns(self.types, columns)

    def drop_rows(self, *keys: str) -> None:
        self.rows = [row for row in self.rows if row[self.key_column] not in keys]

    def _create_ai_columns(self,column_name:str) -> None:
        self.types[column_name] = 'text'
        for x in self.rows:
            x[column_name] = ''

    def _pick_model(self,model_name:str) -> str:
        model_map = {
            "vit-gpt2":"https://api-inference.huggingface.co/models/nlpconnect/vit-gpt2-image-captioning",
            "blip-image":"https://api-inference.huggingface.co/models/Salesforce/blip-image-captioning-base",
            "git-large":"https://api-inference.huggingface.co/models/microsoft/git-large"
        }

        if model_name:
            try:
                model_url = model_map[model_name]
                logger.info(f"Using model {model_name}")
            except:
                logger.error(f"{model_name} is not present! defaulting to vit-gpt2")
                model_url = model_map["vit-gpt2"]
        else:
            logger.error(f"model not provided! defaulting to vit-gpt2")
            model_url = model_map["vit-gpt2"]
        
        return model_url
    
    def _column_types(self, column_types: Optional[List[str]] = None) -> Dict[str, str]:
        
        if not column_types:
            if self.key_column:
                return {
                key: guess_type_by_values(self.col_values(key))
                for key in self.content_columns
            }
            else:
                return {
                    key: guess_type_by_values(self.col_values(key))
                    for key in self.columns[1:]
                }   

        if len(column_types) != len(self.columns) - 1:
            raise CriticalError(
                "Each column (except key) type must be defined in custom types list"
            )

        return {key: column_types[i] for i, key in enumerate(self.content_columns)}



import hashlib
import json
import re
import uuid
import time

from requests import Session, HTTPError
from requests.cookies import cookiejar_from_dict
from urllib.parse import urljoin
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from getpass import getpass

from .block import Block, BLOCK_TYPES
from .collection import (
    Collection,
    CollectionView,
    CollectionRowBlock,
    COLLECTION_VIEW_TYPES,
    TemplateBlock,
)

import logging
from .monitor import Monitor
from .operations import operation_update_last_edited, build_operation
from .settings import API_BASE_URL
from .space import Space
from .store import RecordStore
from .user import User
from .utils import extract_id, now

from icecream import ic
logger = logging.getLogger(__name__)

def create_session(client_specified_retry=None):
    """
    retry on 502
    """
    session = Session()
    if client_specified_retry:
        retry = client_specified_retry
    else:
        retry = Retry(
            5,
            backoff_factor=0.3,
            status_forcelist=(502, 503, 504),
            # CAUTION: adding 'POST' to this list which is not technically idempotent
            allowed_methods=(
                "POST",
                "HEAD",
                "TRACE",
                "GET",
                "PUT",
                "OPTIONS",
                "DELETE",
            ),
        )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("https://", adapter)
    return session


class NotionClient(object):
    """
    This is the entry point to using the API. Create an instance of this class, passing it the value of the
    "token_v2" cookie from a logged-in browser session on Notion.so. Most of the methods on here are primarily
    for internal use -- the main one you'll likely want to use is `get_block`.
    """

    def __init__(
        self,
        token_v2=None,
        monitor=False,
        start_monitoring=False,
        enable_caching=False,
        cache_key=None,
        email=None,
        password=None,
        client_specified_retry=None,
        workspace=None
    ):
        self.workspace = workspace
        self.session = create_session(client_specified_retry)
        if token_v2:
            self.session.cookies = cookiejar_from_dict({"token_v2": token_v2})
        else:
            self._set_token(email=email, password=password)

        if enable_caching:
            cache_key = cache_key or hashlib.sha256(token_v2.encode()).hexdigest()
            self._store = RecordStore(self, cache_key=cache_key)
        else:
            self._store = RecordStore(self)
        if monitor:
            self._monitor = Monitor(self)
            if start_monitoring:
                self.start_monitoring()
        else:
            self._monitor = None

        self._update_user_info()

    def start_monitoring(self):
        self._monitor.poll_async()
    
    def _fetch_guest_space_data(self, records):
        """
        guest users have an empty `space` dict, so get the space_id from the `space_view` dict instead,
        and fetch the space data from the getPublicSpaceData endpoint.

        Note: This mutates the records dict
        """
        space_id = list(records["space_view"].values())[0]["value"]["space_id"]

        space_data = self.post(
            "getPublicSpaceData", {"type": "space-ids", "spaceIds": [space_id]}
        ).json()

        records["space"] = {
            space["id"]: {"value": space} for space in space_data["results"]
        }


    def _set_token(self, email=None, password=None):
        if not email:
            email = input("Enter your Notion email address:\n")
        if not password:
            password = getpass("Enter your Notion password:\n")
        self.post("loginWithEmail", {"email": email, "password": password}).json()

    def _update_user_info(self):
        records = self.post("loadUserContent", {}).json()["recordMap"]

        if not records["space"]:
            self._fetch_guest_space_data(records)

        if self.workspace:
            workspace_dict = {}
            for x in records['space']:
                workspace_dict[records['space'][x]['value']['name']] = x
                
                try:
                    self.current_space = self.get_space(workspace_dict[self.workspace])
                except:
                    self.current_space = self.get_space(list(records["space"].keys())[0])
                    logger.warning(f"Workspace {self.workspace} does not exist or invalid, defaulting to {self.current_space.name}")

        else:
            self.current_space = self.get_space(list(records["space"].keys())[0])
            logger.warning(f"Workspace not provided or invalid, defaulting to {self.current_space.name}")

        self._store.store_recordmap(records)
        self.current_user = self.get_user(list(records["notion_user"].keys())[0])
        # self.current_space = self.get_space(list(records["space"].keys())[0])

        self.session.headers.update(
            {"x-notion-active-user-header": self.current_user.id}
        )

        return records

    def get_email_uid(self):
        response = self.post("getSpaces", {}).json()
        return {
            response[uid]["notion_user"][uid]["value"]["email"]: uid
            for uid in response.keys()
        }

    def set_user_by_uid(self, user_id):
        self.session.headers.update({"x-notion-active-user-header": user_id})
        self._update_user_info()

    def set_user_by_email(self, email):
        email_uid_dict = self.get_email_uid()
        uid = email_uid_dict.get(email)
        if not uid:
            raise Exception(
                "Requested email address {email} not found; available addresses: {available}".format(
                    email=email, available=list(email_uid_dict)
                )
            )
        self.set_user_by_uid(uid)

    def get_top_level_pages(self):
        records = self._update_user_info()
        return [self.get_block(bid) for bid in records["block"].keys()]

    def get_record_data(self, table, id, force_refresh=False, limit=100):
        return self._store.get(table, id, force_refresh=force_refresh, limit=limit)

    def get_block(self, url_or_id, force_refresh=False, limit=100):
        """
        Retrieve an instance of a subclass of Block that maps to the block/page identified by the URL or ID passed in.
        """
        block_id = extract_id(url_or_id)
        block = self.get_record_data("block", block_id, force_refresh=force_refresh, limit=limit)
        if not block:
            return None
        if block.get("parent_table") == "collection":
            if block.get("is_template"):
                block_class = TemplateBlock
            else:
                block_class = CollectionRowBlock
        else:
            block_class = BLOCK_TYPES.get(block.get("type", ""), Block)
        return block_class(self, block_id)

    def get_collection(self, collection_id, force_refresh=False):
        """
        Retrieve an instance of Collection that maps to the collection identified by the ID passed in.
        """
        coll = self.get_record_data(
            "collection", collection_id, force_refresh=force_refresh
        )
        return Collection(self, collection_id) if coll else None

    def get_user(self, user_id, force_refresh=False):
        """
        Retrieve an instance of User that maps to the notion_user identified by the ID passed in.
        """
        user = self.get_record_data("notion_user", user_id, force_refresh=force_refresh)
        return User(self, user_id) if user else None

    def get_space(self, space_id, force_refresh=False):
        """
        Retrieve an instance of Space that maps to the space identified by the ID passed in.
        """
        space = self.get_record_data("space", space_id, force_refresh=force_refresh)
        return Space(self, space_id) if space else None

    def get_collection_view(self, url_or_id, collection=None, force_refresh=False):
        """
        Retrieve an instance of a subclass of CollectionView that maps to the appropriate type.
        The `url_or_id` argument can either be the URL for a database page, or the ID of a collection_view (in which case
        you must also pass the collection)
        """
        # if it's a URL for a database page, try extracting the collection and view IDs
        if url_or_id.startswith("http"):
            match = re.search(r"([a-f0-9]{32})\?v=([a-f0-9]{32})", url_or_id)
            if not match:
                raise Exception("Invalid collection view URL")
            block_id, view_id = match.groups()
            collection = self.get_block(
                block_id, force_refresh=force_refresh
            ).collection
        else:
            view_id = url_or_id
            assert (
                collection is not None
            ), "If 'url_or_id' is an ID (not a URL), you must also pass the 'collection'"

        view = self.get_record_data(
            "collection_view", view_id, force_refresh=force_refresh
        )

        return (
            COLLECTION_VIEW_TYPES.get(view.get("type", ""), CollectionView)(
                self, view_id, collection=collection
            )
            if view
            else None
        )

    def refresh_records(self, **kwargs):
        """
        The keyword arguments map table names into lists of (or singular) record IDs to load for that table.
        Use `True` instead of a list to refresh all known records for that table.
        """
        self._store.call_get_record_values(**kwargs)

    def refresh_collection_rows(self, collection_id):
        row_ids = [row.id for row in self.get_collection(collection_id).get_rows()]
        self._store.set_collection_rows(collection_id, row_ids)

    def post(self, endpoint, data):
        """
        All API requests on Notion.so are done as POSTs (except the websocket communications).
        """
        url = urljoin(API_BASE_URL, endpoint)

        # Retry logic has been implemented, maximum retry is 6 with each retry waiting for 1(base time) * (1.5 ^ retry_count), the wait time increase exponentially in terms of 1.5 ^ n and maximum wait time is 1.5 ^ 9 = 38 sec
        # ie, for the first retry count, the sleep time is 1 * 1.5 = 1.5 sec. for 2nd it is 1 * (1.5 ^ 2) = 2.25 sec and so on. 
        retry_count = 0
        retry_limit = 15

        while True:
            
            if retry_count > retry_limit:
                break

            #print(retry_count)
            response = self.session.post(url, json=data)
            if response.status_code == 400:
                if endpoint != 'getRecordValues':
                    logger.error(
                        "Got 400 error attempting to POST to {}, with data: {}".format(
                            endpoint, json.dumps(data, indent=2)
                        )
                    )
                    raise HTTPError(
                        response.json().get(
                            "message", "There was an error (400) submitting the request."
                        )
                    )
            
            # Only retry if its an internal server error (ie 500)
            if response.status_code == 500:
                retry_count += 1
            else:
                break
            
            # Exponential backoff wait time
            wait_time = (1.5 ** retry_count) 
            time.sleep(wait_time)
            
        response.raise_for_status()
        return response

    def submit_transaction(self, operations, update_last_edited=True):
        
        if not operations:
            return

        if isinstance(operations, dict):
            operations = [operations]

        if update_last_edited:
            updated_blocks = set(
                [op["id"] for op in operations if op["table"] == "block"]
            )
            operations += [
                operation_update_last_edited(self.current_user.id, block_id)
                for block_id in updated_blocks
            ]

        # if we're in a transaction, just add these operations to the list; otherwise, execute them right away
        if self.in_transaction():
            self._transaction_operations += operations
        else:
            data = {"operations": operations}
            
            self.post("submitTransaction", data)
            self._store.run_local_operations(operations)

    def query_collection(self, *args, **kwargs):
        return self._store.call_query_collection(*args, **kwargs)

    def as_atomic_transaction(self):
        """
        Returns a context manager that buffers up all calls to `submit_transaction` and sends them as one big transaction
        when the context manager exits.
        """
        return Transaction(client=self)

    def in_transaction(self):
        """
        Returns True if we're currently in a transaction, otherwise False.
        """
        return hasattr(self, "_transaction_operations")

    def search_pages_with_parent(self, parent_id, search="", limit=100):
        data = {
            "query": search,
            "parentId": parent_id,
            "limit": limit,
            "spaceId": self.current_space.id,
        }
        response = self.post("searchPagesWithParent", data).json()
        self._store.store_recordmap(response["recordMap"])
        return response["results"]

    def search_blocks(self, search, limit=25):
        return self.search(query=search, limit=limit)

    def search(
        self,
        query="",
        search_type="BlocksInSpace",
        limit=100,
        sort="Relevance",
        source="quick_find",
        isDeletedOnly=False,
        excludeTemplates=False,
        isNavigableOnly=False,
        requireEditPermissions=False,
        ancestors=[],
        createdBy=[],
        editedBy=[],
        lastEditedTime={},
        createdTime={},
    ):
        data = {
            "type": search_type,
            "query": query,
            "spaceId": self.current_space.id,
            "limit": limit,
            "filters": {
                "isDeletedOnly": isDeletedOnly,
                "excludeTemplates": excludeTemplates,
                "isNavigableOnly": isNavigableOnly,
                "requireEditPermissions": requireEditPermissions,
                "ancestors": ancestors,
                "createdBy": createdBy,
                "editedBy": editedBy,
                "lastEditedTime": lastEditedTime,
                "createdTime": createdTime,
            },
            "sort": sort,
            "source": source,
        }
        response = self.post("search", data).json()
        self._store.store_recordmap(response["recordMap"])
        return [self.get_block(result["id"]) for result in response["results"]]

    def create_record(self, table, parent, **kwargs):

        # make up a new UUID; apparently we get to choose our own!
        record_id = str(uuid.uuid4())

        child_list_key = kwargs.get("child_list_key") or parent.child_list_key

        args = {
            "id": record_id,
            "version": 1,
            "alive": True,
            "created_by_id": self.current_user.id,
            "created_by_table": "notion_user",
            "created_time": now(),
            "parent_id": parent.id,
            "parent_table": parent._table,
            "space_id":self.current_space.id #this is required to create a CRDT metadata by notion
        }

        args.update(kwargs)

        with self.as_atomic_transaction():
            # create the new record 
            self.submit_transaction(
                build_operation(
                    args=args, command="set", id=record_id, path=[], table=table
                )
            )

            # add the record to the content list of the parent, if needed
            if child_list_key:
                self.submit_transaction(
                    build_operation(
                        id=parent.id,
                        path=[child_list_key],
                        args={"id": record_id},
                        command="listAfter",
                        table=parent._table,
                    )
                )

        return record_id


class Transaction(object):

    
    is_dummy_nested_transaction = False

    def __init__(self, client):
        self.client = client

    def __enter__(self):
        
        if hasattr(self.client, "_transaction_operations"):
            # client is already in a transaction, so we'll just make this one a nullop and let the outer one handle it
            self.is_dummy_nested_transaction = True
            return

        self.client._transaction_operations = []
        self.client._pages_to_refresh = []
        self.client._blocks_to_refresh = []

    def __exit__(self, exc_type, exc_value, traceback):

        if self.is_dummy_nested_transaction:
            return

        operations = self.client._transaction_operations
        del self.client._transaction_operations

        # only actually submit the transaction if there was no exception
        if not exc_type:
            self.client.submit_transaction(operations)

        self.client._store.handle_post_transaction_refreshing()



import json
import re
import requests
import threading
import time
import uuid

from collections import defaultdict
from inspect import signature
from requests import HTTPError

from .collection import Collection
from .logger import logger
from .records import Record


class Monitor(object):

    thread = None

    def __init__(self, client, root_url="https://msgstore.www.notion.so/primus/"):
        self.client = client
        self.session_id = str(uuid.uuid4())
        self.root_url = root_url
        self._subscriptions = set()
        self.initialize()

    def _decode_numbered_json_thing(self, thing):

        thing = thing.decode().strip()

        for ping in re.findall('\d+:\d+"primus::ping::\d+"', thing):
            logger.debug("Received ping: {}".format(ping))
            self.post_data(ping.replace("::ping::", "::pong::"))

        results = []
        for blob in re.findall("\d+:\d+(\{.*?\})(?=\d|$)", thing):
            results.append(json.loads(blob))
        if thing and not results and "::ping::" not in thing:
            logger.debug("Could not parse monitoring response: {}".format(thing))
        return results

    def _encode_numbered_json_thing(self, data):
        assert isinstance(data, list)
        results = ""
        for obj in data:
            msg = str(len(obj)) + json.dumps(obj, separators=(",", ":"))
            msg = "{}:{}".format(len(msg), msg)
            results += msg
        return results.encode()

    def initialize(self):

        logger.debug("Initializing new monitoring session.")

        response = self.client.session.get(
            "{}?sessionId={}&EIO=3&transport=polling".format(
                self.root_url, self.session_id
            )
        )

        self.sid = self._decode_numbered_json_thing(response.content)[0]["sid"]

        logger.debug("New monitoring session ID is: {}".format(self.sid))

        # resubscribe to any existing subscriptions if we're reconnecting
        old_subscriptions, self._subscriptions = self._subscriptions, set()
        self.subscribe(old_subscriptions)

    def subscribe(self, records):

        if isinstance(records, set):
            records = list(records)

        if not isinstance(records, list):
            records = [records]

        sub_data = []

        for record in records:

            if record not in self._subscriptions:

                logger.debug(
                    "Subscribing new record to the monitoring watchlist: {}/{}".format(
                        record._table, record.id
                    )
                )

                # add the record to the list of records to restore if we're disconnected
                self._subscriptions.add(record)

                # subscribe to changes to the record itself
                sub_data.append(
                    {
                        "type": "/api/v1/registerSubscription",
                        "requestId": str(uuid.uuid4()),
                        "key": "versions/{}:{}".format(record.id, record._table),
                        "version": record.get("version", -1),
                    }
                )

                # if it's a collection, subscribe to changes to its children too
                if isinstance(record, Collection):
                    sub_data.append(
                        {
                            "type": "/api/v1/registerSubscription",
                            "requestId": str(uuid.uuid4()),
                            "key": "collection/{}".format(record.id),
                            "version": -1,
                        }
                    )

        data = self._encode_numbered_json_thing(sub_data)

        self.post_data(data)

    def post_data(self, data):

        if not data:
            return

        logger.debug("Posting monitoring data: {}".format(data))

        self.client.session.post(
            "{}?sessionId={}&transport=polling&sid={}".format(
                self.root_url, self.session_id, self.sid
            ),
            data=data,
        )

    def poll(self, retries=10):
        logger.debug("Starting new long-poll request")
        try:
            response = self.client.session.get(
                "{}?sessionId={}&EIO=3&transport=polling&sid={}".format(
                    self.root_url, self.session_id, self.sid
                )
            )
            response.raise_for_status()
        except HTTPError as e:
            try:
                message = "{} / {}".format(response.content, e)
            except:
                message = "{}".format(e)
            logger.warn(
                "Problem with submitting polling request: {} (will retry {} more times)".format(
                    message, retries
                )
            )
            time.sleep(0.1)
            if retries <= 0:
                raise
            if retries <= 5:
                logger.error(
                    "Persistent error submitting polling request: {} (will retry {} more times)".format(
                        message, retries
                    )
                )
                # if we're close to giving up, also try reinitializing the session
                self.initialize()
            self.poll(retries=retries - 1)

        self._refresh_updated_records(
            self._decode_numbered_json_thing(response.content)
        )

    def _refresh_updated_records(self, events):

        records_to_refresh = defaultdict(list)

        for event in events:

            logger.debug(
                "Received the following event from the remote server: {}".format(event)
            )

            if not isinstance(event, dict):
                continue

            if event.get("type", "") == "notification":

                key = event.get("key")

                if key.startswith("versions/"):

                    match = re.match(r"versions/([^\:]+):(.+)", key)
                    if not match:
                        continue

                    record_id, record_table = match.groups()

                    local_version = self.client._store.get_current_version(
                        record_table, record_id
                    )
                    if event["value"] > local_version:
                        logger.debug(
                            "Record {}/{} has changed; refreshing to update from version {} to version {}".format(
                                record_table, record_id, local_version, event["value"]
                            )
                        )
                        records_to_refresh[record_table].append(record_id)
                    else:
                        logger.debug(
                            "Record {}/{} already at version {}, not trying to update to version {}".format(
                                record_table, record_id, local_version, event["value"]
                            )
                        )

                if key.startswith("collection/"):

                    match = re.match("collection/(.+)", key)
                    if not match:
                        continue

                    collection_id = match.groups()[0]

                    self.client.refresh_collection_rows(collection_id)
                    row_ids = self.client._store.get_collection_rows(collection_id)

                    logger.debug(
                        "Something inside collection {} has changed; refreshing all {} rows inside it".format(
                            collection_id, len(row_ids)
                        )
                    )

                    records_to_refresh["block"] += row_ids

        self.client.refresh_records(**records_to_refresh)

    def poll_async(self):
        if self.thread:
            # Already polling async; no need to have two threads
            return
        self.thread = threading.Thread(target=self.poll_forever, daemon=True)
        self.thread.start()

    def poll_forever(self):
        while True:
            try:
                self.poll()
            except Exception as e:
                logger.error("Encountered error during polling!")
                logger.error(e, exc_info=True)
                time.sleep(1)



import shutil, os
import logging
from tqdm import tqdm
from icecream import ic
import requests
import time

logger = logging.getLogger(__name__)

class AI:

    def __init__(self,ai_data:dict) -> None:
        self.ai_data = ai_data
    
    def out(self,row) -> str:
        
        if 'caption' in self.ai_data:
            data = self.ai_data['caption']
            try:
                row[data['caption_column']] = self._img2caption(data['hftoken'],str(data['image_path']),data['model_url'])
                return row
            except Exception as e:
                logger.error(f"Error during AI process : {e}")


    def _img2caption(self,token:str,image_url:str,model_url:str) -> str:
        try:
            file = open(image_url,'rb')
            filename = os.path.basename(image_url)
            tqdm.write(f"AI generating caption for image {filename}")

            retries = 0
            while True:

                if retries == 15:
                    tqdm.write(f"Error generating caption for {filename} \n The model size is huge and could have problem loading! try again after some time")
                    #logger.error(caption.json())
                    break

                sess = requests.session()
                caption = sess.post(
                model_url,
                json='None',
                data=file,
                headers={'authorization': f'Bearer {token}'},
                cookies=None,
                timeout=None,
                stream=False
                )

                retries += 1
                if 'error' in caption.json():
                    if 'estimated_time' in caption.json():
                        time.sleep(3)
                    elif 'Error in `parameters`: field required' in caption.json()['error']:
                        time.sleep(3)
                    else:
                        break
                else:
                    break

            tqdm.write(f"Caption generated for image {filename} : {caption.json()[0]['generated_text']}")
            return caption.json()[0]['generated_text']
        except Exception as e:
            tqdm.write(f"Error generating caption for {filename}")
            logger.error(e,exc_info=1)
            logger.error(caption.json())




        



import math
import re
from typing import List
from icecream import ic

def guess_type_by_values(values_str: List[str]) -> str:

    if type(values_str[0]) == list:
        unique_values = set(map(tuple, values_str))
    else:   
        unique_values = set(filter(None, values_str))

    match_map = {
        "text": is_empty,
        "checkbox": is_checkbox,
        "number": is_number,
        "url": is_url,
        "email": is_email,
        "multi_select": is_multi
    }

    matches = (
        value_type
        for value_type, match_func in match_map.items()
        if all(map(match_func, unique_values))
    )
           
    return next(matches, "text")


def is_number(s: str) -> bool:
    try:
        return not math.isnan(float(s))
    except:
        return False

def is_multi(s) -> bool:
    
    try:
        if type(s) == tuple:
            return True
    except:
        return None
    
def is_url(s: str) -> bool:

    try:
        return re.match("^https?://", s) is not None
    except:
        return None


def is_email(s: str) -> bool:
    try:
        return re.match(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$", s) is not None
    except:
        return None


def is_checkbox(s: str) -> bool:
    try:
        return s in {"true", "false"}
    except:
        return None


def is_empty(s: str) -> bool:

    try:
        return not s.strip()
    except:
        return None



import logging
import os

from .settings import LOG_FILE


NOTIONPY_LOG_LEVEL = os.environ.get("NOTIONPY_LOG_LEVEL", "warning").lower()

logger = logging.getLogger("notion")


def enable_debugging():
    set_log_level(logging.DEBUG)


def set_log_level(level):
    logger.setLevel(level)
    handler.setLevel(level)


if NOTIONPY_LOG_LEVEL == "disabled":
    handler = logging.NullHandler()
    logger.addHandler(handler)
else:
    handler = logging.FileHandler(LOG_FILE)
    formatter = logging.Formatter("\n%(asctime)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    if NOTIONPY_LOG_LEVEL == "debug":
        set_log_level(logging.DEBUG)
    elif NOTIONPY_LOG_LEVEL == "info":
        set_log_level(logging.INFO)
    elif NOTIONPY_LOG_LEVEL == "warning":
        set_log_level(logging.WARNING)
    elif NOTIONPY_LOG_LEVEL == "error":
        set_log_level(logging.ERROR)
    else:
        raise Exception(
            "Invalid value for environment variable NOTIONPY_LOG_LEVEL: {}".format(
                NOTIONPY_LOG_LEVEL
            )
        )



import hashlib
from pathlib import Path


def get_file_sha256(file_path: Path) -> str:
    hash_sha256 = hashlib.sha256()
    chunk_size = 4096

    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):  # noqa: WPS426
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()



import mimetypes
import re
from pathlib import Path
from typing import Dict, Optional, Tuple

import requests
from csv2notion_neo.notion.block import Block

from csv2notion_neo.utils_exceptions import NotionError
from csv2notion_neo.utils_file import get_file_sha256
from csv2notion_neo.utils_static import FileType
from urllib.parse import urlparse
from icecream import ic

Meta = Dict[str, str]


def upload_filetype(parent: Block, filetype: FileType) -> Tuple[str, Meta]:

    if isinstance(filetype, Path):
        url, meta = upload_file(parent, filetype)
    else:
        url = filetype
        meta = {
        "source": [
            [
                url
            ]
        ]
    }

    return url, meta


def upload_file(block: Block, file_path: Path) -> Tuple[str, Meta]:

        
    file_url = _upload_file(block, file_path)

    file_id = get_file_id(file_url)

    if file_id is None:
        raise NotionError(f"Could not upload file {file_path}")
    
    return file_url, {
        "source": [
            [
                file_url
            ]
        ]
    }
    # return file_url, {
    #     "type": "file",
    #     "file_id": file_id,
    #     "sha256": get_file_sha256(file_path),
    # }


def _upload_file(block: Block, file_path: Path) -> str:
    file_mime = mimetypes.guess_type(file_path.name)[0] or "application/octet-stream"

    post_data = {
        "bucket": "secure",
        "name": file_path.name,
        "contentType": file_mime,
        "record": {
            "table": "block",
            "id": block.id,
            "spaceId": block.space_info["spaceId"],
        },
    }

    upload_data = block._client.post("getUploadFileUrl", post_data).json()

    
    with open(file_path, "rb") as f:
        requests.put(
            upload_data["signedPutUrl"],
            data=f,
            headers={"Content-type": file_mime},
        ).raise_for_status()

    return str(upload_data.get("url", ""))


def get_file_id(image_url: str) -> Optional[str]:

    
    # aws_host/space_id/file_id/filename
    #aws_re = r"^https://(.*?\.amazonaws\.com)/([a-f0-9\-]+)/([a-f0-9\-]+)/(.*?)$"
    attachment_re = r"^attachment:([a-f0-9\-]+):(.+)$"
    aws_match = re.search(attachment_re, image_url)

    if aws_match:
        return aws_match.group(1)

    return None


def is_meta_different(
    image: Optional[FileType],
    image_url: Optional[str],
    image_meta: Optional[Meta],
) -> bool:
    if not all([image_meta, image_url, image]):
        return True

    if isinstance(image, Path) and image_url is not None and image_meta is not None:
        return _is_file_meta_different(image, image_url, image_meta)

    elif image_meta != {"type": "url", "url": image}:
        return True

    return False


def _is_file_meta_different(image: Path, image_url: str, image_meta: Meta) -> bool:
    if image_meta["type"] != "file":
        return True

    if image_meta["file_id"] != get_file_id(image_url):
        return True

    return image_meta["sha256"] != get_file_sha256(image)



from datetime import datetime
from pathlib import Path
from typing import Optional, Union

from dateutil.parser import ParserError
from dateutil.parser import parse as date_parse
from emoji import distinct_emoji_list, emoji_count, replace_emoji
from csv2notion_neo.notion.collection import NotionDate

from csv2notion_neo.notion_type_guess import is_url
from csv2notion_neo.utils_exceptions import TypeConversionError
from csv2notion_neo.utils_static import FileType
from csv2notion_neo.utils_str import split_str
from icecream import ic

def map_checkbox(s: str) -> bool:
    return s == "true"


def map_date(s: str) -> datetime:
    try:
        return date_parse(s)
    except ParserError as e:
        raise TypeConversionError(e) from e


def map_notion_date(s: str) -> NotionDate:

    dates = split_str(s)

    if not dates:
        raise TypeConversionError("Date field is empty")

    if len(dates) > 2:
        raise TypeConversionError("Date field doesn't support more than 2 values")

    if len(dates) == 2:
        return NotionDate(start=map_date(dates[0]), end=map_date(dates[1]))

    return NotionDate(start=map_date(dates[0]))


def map_number(s: str) -> Union[int, float]:
    try:
        float_value = float(s)
    except ValueError as e:
        raise TypeConversionError(e) from e

    if float_value.is_integer():
        return int(float_value)

    return float_value


def map_icon(s: str) -> FileType:
    icon_emoji = _get_icon_emoji(s)
    if icon_emoji:
        return icon_emoji

    return s if is_url(s) else Path(s)


def map_url_or_file(s: str) -> FileType:
    return s if is_url(s) else Path(s)


def _get_icon_emoji(s: str) -> Optional[str]:
    # string has anything other than emoji
    if replace_emoji(s) != "":
        return None

    # string must contain exactly one emoji for icon
    if emoji_count(s) != 1:
        return None

    return str(distinct_emoji_list(s)[0])



from datetime import datetime
from itertools import starmap
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple


from csv2notion_neo.utils_exceptions import NotionError
from cached_property import cached_property
from csv2notion_neo.notion.collection import CollectionRowBlock
from csv2notion_neo.notion.maps import field_map
from csv2notion_neo.notion.operations import build_operation
from csv2notion_neo.notion.utils import remove_signed_prefix_as_needed

from csv2notion_neo.utils_exceptions import NotionError
from csv2notion_neo.notion_row_image_block import RowCoverImageBlock
from csv2notion_neo.notion_row_upload_file import Meta, is_meta_different, upload_filetype
from csv2notion_neo.utils_static import FileType
from icecream import ic
from csv2notion_neo.notion_row_upload_file import get_file_id

NamedURLs = Dict[str, str]

class CollectionRowBlockExtended(CollectionRowBlock):  # noqa: WPS214
    icon_meta = field_map("properties.meta.icon")
    cover_meta = field_map("properties.meta.cover")
    cover_block_meta = field_map("properties.meta.cover_block")

    @cached_property  # type: ignore
    def image_block(self) -> RowCoverImageBlock:
        return RowCoverImageBlock(self)

    @property
    def icon(self) -> str:
        return super().icon  # type: ignore

    @icon.setter
    def icon(self, icon: FileType) -> None:
        new_icon: Optional[FileType] = icon if icon else None

        if not self._is_meta_changed("icon_meta", new_icon, self.icon):
            return

        if new_icon is None:
            icon_meta = None
        else:
            new_icon, icon_meta = upload_filetype(self, new_icon)

        self.icon_meta = icon_meta
        CollectionRowBlock.icon.fset(self, new_icon)

    @property
    def cover(self) -> Optional[str]:
        return super().cover  # type: ignore
        #return self.image_block.url  # type: ignore

    @cover.setter
    def cover(self, image: FileType) -> None:   
    
        if image:
            cover_img = image.pop(0)
        else:
            cover_img = None

        if self._client.in_transaction():
            raise RuntimeError("Cannot set cover_block during atomic transaction")
        
        new_image: Optional[FileType] = cover_img if cover_img else None

        if not self._is_meta_changed("cover_meta", new_image, self.cover):
            return

        if new_image is None:
            cover_meta = None
        else:
            new_image, cover_meta = upload_filetype(self, new_image)

        self.cover_meta = cover_meta
        CollectionRowBlock.cover.fset(self, new_image)

        #new_images.append(new_image)

        #self.image_block.url = new_images

        
    @property
    def cover_block(self) -> Optional[str]:
        return self.image_block.url  # type: ignore

    @cover_block.setter
    def cover_block(self, image: FileType) -> None:
        
        new_images = []
        if image:
            cover_img = image.pop(0)
        else:
            cover_img = None
        if self._client.in_transaction():
            raise RuntimeError("Cannot set cover_block during atomic transaction")

        new_image: Optional[FileType] = cover_img if cover_img else None
     
        if not self._is_meta_changed("cover_block_meta", new_image, self.cover_block):
            return

        if new_image is None:
            cover_block_meta = None
        else:
            new_image, cover_block_meta = upload_filetype(self, new_image)


        self.cover_block_meta = cover_block_meta
        new_images.append(new_image)
        

        if image:
            for im in image:
                new_img,meta = upload_filetype(self,im)
                new_images.append(new_img)

        self.image_block.url = new_images

    def upload_non_cover_images(self, image_url: Optional[str]) -> None:
        if image_url is None:
            if self.image_block is not None:
                self.image_block.remove()
            return

        attrs = {
            "display_source": image_url,
            "source": image_url,
        }
        
        file_id = get_file_id(image_url)
        if file_id:
            attrs["file_id"] = file_id

    @property
    def cover_block_caption(self) -> Optional[str]:
        return self.image_block.caption  # type: ignore

    @cover_block_caption.setter
    def cover_block_caption(self, caption: Optional[str]) -> None:
        if self.cover_block_caption == caption:
            return

        self.image_block.caption = caption

    @property
    def created_time(self) -> datetime:
        created_time = self.get("created_time")
        return datetime.utcfromtimestamp(created_time / 1000)

    @created_time.setter
    def created_time(self, time: datetime) -> None:
        self.set("created_time", int(time.timestamp() * 1000))

    @property
    def last_edited_time(self) -> datetime:
        last_edited_time = self.get("last_edited_time")
        return datetime.utcfromtimestamp(last_edited_time / 1000)

    @last_edited_time.setter
    def last_edited_time(self, time: datetime) -> None:
        if self._client.in_transaction():
            raise RuntimeError("Cannot set last_edited_time during atomic transaction")

        self._client.submit_transaction(
            build_operation(
                id=self.id,
                path="last_edited_time",
                args=int(time.timestamp() * 1000),
                table=self._table,
            ),
            update_last_edited=False,
        )

    def set_property(self, identifier: str, new_value: Any) -> None:
        prop = self.collection.get_schema_property(identifier)
        if prop is None:
            raise AttributeError(f"Object does not have property '{identifier}'")

        if prop["type"] in {"select", "multi_select"}:
            schema_update, prop = self.collection.check_schema_select_options(
                prop, new_value
            )
            if schema_update:
                self.collection.set(
                    "schema.{0}.options".format(prop["id"]), prop["options"]
                )

        if prop["type"] == "file":
            if not self._is_file_column_changed(prop["id"], new_value):
                return
            new_value = self._upload_column_files(prop["id"], new_value)

        path, new_value = self._convert_python_to_notion(
            new_value, prop, identifier=identifier
        )

        self.set(path, new_value)

    def _convert_notion_to_python(self, val: Any, prop: Dict[str, Any]) -> Any:
        if prop["type"] == "status":
            return val[0][0] if val else None

        return super()._convert_notion_to_python(val, prop)

    def _convert_python_to_notion(
        self, raw_value: Any, prop: Dict[str, Any], identifier: str = "<unknown>"
    ) -> Any:
        result_value = None

        if prop["type"] == "file" and isinstance(raw_value, dict):
            filelist = []
            for filename, url in raw_value.items():
                url = remove_signed_prefix_as_needed(url)
                filelist += [[filename, [["a", url]]], [","]]
            result_value = filelist[:-1]

        if prop["type"] == "status":
            if not raw_value:
                result_value = [[""]]
            elif isinstance(raw_value, str):
                valid_options = [p["value"].lower() for p in prop["options"]]

                if raw_value.lower() not in valid_options:
                    raise ValueError(
                        f"Value '{raw_value}' not acceptable for property"
                        f" '{identifier}' (valid options: {valid_options})"
                    )
                result_value = [[raw_value]]
            else:
                raise ValueError(
                    f"Wrong value type for 'status' type column '{identifier}'"
                )

        if result_value is not None:
            return ["properties", prop["id"]], result_value
        
        try:
            return super()._convert_python_to_notion(raw_value, prop, identifier)
        except:
            raise NotionError(f"Column {prop['name']} seems to have an property mismatch, please check and change the property to appropriate value in notion database!")
        

    def _upload_column_files(self, column_id: str, files: List[FileType]) -> NamedURLs:
        
        column_files_meta, column_files_urls = self._process_column_files(files)

        self.set(f"properties.meta.file_columns.{column_id}", column_files_meta)

        return column_files_urls

    def _process_column_files(
        self, files: List[FileType]
    ) -> Tuple[List[Meta], NamedURLs]:
        
       
        column_files_meta: List[Meta] = []
        column_files_urls: NamedURLs = {}

        for filetype in files:
            file_url, file_meta = upload_filetype(self, filetype)

            column_files_urls[get_filetype_name(filetype)] = file_url
            column_files_meta.append(file_meta)

        return column_files_meta, column_files_urls

    def _is_file_column_changed(self, column_id: str, files: List[FileType]) -> bool:
        files_meta = self.get(f"properties.meta.file_columns.{column_id}")
        files_url = self.get_property(column_id)

        # if column is empty and files list is empty, nothing to change
        if not any([files, files_meta, files_url]):
            return False

        if not files_meta or not files_url:
            return True

        if not (len(files_meta) == len(files_url) == len(files)):  # noqa: WPS508
            return True

        return any(starmap(is_meta_different, zip(files, files_url, files_meta)))

    def _is_meta_changed(
        self,
        meta_parameter: str,
        new_val: Optional[FileType],
        current_val: Optional[str],
    ) -> bool:
        if (current_val is None or new_val is None) and current_val != new_val:
            return True

        current_value_meta = getattr(self, meta_parameter)


        return is_meta_different(new_val, current_val, current_value_meta)


def get_filetype_name(filetype: FileType) -> str:
    return filetype.name if isinstance(filetype, Path) else filetype



import logging
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from csv2notion_neo.notion.user import User
from csv2notion_neo.notion.utils import InvalidNotionIdentifier, extract_id

from csv2notion_neo.local_data import LocalData, CSVRowType
from csv2notion_neo.notion_convert_map import (
    map_checkbox,
    map_date,
    map_icon,
    map_notion_date,
    map_number,
    map_url_or_file,
)
from csv2notion_neo.notion_db import NotionDB
from csv2notion_neo.notion_row import CollectionRowBlockExtended
from csv2notion_neo.notion_type_guess import is_email
from csv2notion_neo.notion_uploader import NotionUploadRow
from csv2notion_neo.utils_exceptions import NotionError, TypeConversionError, CriticalError
from csv2notion_neo.utils_static import ConversionRules, FileType
from csv2notion_neo.utils_str import split_str
from icecream import ic

logger = logging.getLogger(__name__)


class NotionRowConverter(object):  # noqa:  WPS214
    def __init__(self, db: NotionDB, conversion_rules: ConversionRules):
        self.db = db
        self.rules = conversion_rules
        self._current_row = 0

    def convert_to_notion_rows(self, csv_data: LocalData) -> List[NotionUploadRow]:
        notion_rows = []
        # starting with 2nd row, because first is header
        self._current_row = 2
        self.csv_data = csv_data

        for row in csv_data:

            if self.rules.rename_notion_key_column:
                if self.rules.rename_notion_key_column[1] == self.rules.rename_notion_key_column[0]:
                    raise CriticalError(f"Please do not provide same column name in rename-payload-key-column")
                new_id = self.rules.rename_notion_key_column[1]
                old_id = self.rules.rename_notion_key_column[0]
                row[new_id] = row[old_id]
                del row[old_id]

            try:
                notion_rows.append(self._convert_row(row))
            except NotionError as e:
                raise NotionError(f"CSV [{self._current_row}]: {e}")
            self._current_row += 1

        return notion_rows

    def _error(self, error: str) -> None:
        logger.error(f"CSV [{self._current_row}]: {error}")

        if self.rules.fail_on_conversion_error:
            raise NotionError("Error during conversion.")

    def _convert_row(self, row: CSVRowType) -> NotionUploadRow:
        properties = self._map_properties(row)
        columns = self._map_columns(row)
        
        return NotionUploadRow(columns=columns, properties=properties)

    def _map_properties(self, row: CSVRowType) -> Dict[str, Any]:
        properties = {}

        if self.rules.payload_key_column:
            if self.rules.rename_notion_key_column:
                new_id = self.rules.rename_notion_key_column[1]
                properties["payload_key_column"] = new_id
            else:
                properties["payload_key_column"] = self.rules.payload_key_column

        #ai features
        if self.rules.hugging_face_token:
            if self.rules.caption_column:
                properties['AI'] = {"caption":{
                    "hftoken":self.rules.hugging_face_token,
                    "image_path": self.rules.csv_file.parent / Path(row[self.rules.caption_column[0]]),
                    "caption_column":self.rules.caption_column[1],
                    "model_url":self.csv_data.model_url
                }}

        if self.rules.image_column_mode == "block":
            properties["cover_block"] = self._map_image(row)
            properties["cover_block_caption"] = self._map_image_caption(row)
        else:
            # properties["cover_block"] = self._map_image(row)
            # properties["cover_block_caption"] = self._map_image_caption(row)
            properties["cover"] = self._map_image(row)

        properties["icon"] = self._map_icon(row)

        properties["created_time"] = self._pop_column_type(row, "created_time")
        properties["last_edited_time"] = self._pop_column_type(row, "last_edited_time")

        return {k: v for k, v in properties.items() if v is not None}

    def _map_columns(self, row: CSVRowType) -> Dict[str, Any]:
        notion_row = {}

        for col_key, col_value in row.items():

            col_type = self.db.columns[col_key]["type"]
            
            notion_row[col_key] = self._map_column(col_key, col_value, col_type)

            self._raise_if_mandatory_empty(col_key, notion_row[col_key])

        return notion_row

    def _map_column(
        self, col_key: str, col_value: str, value_type: str
    ) -> Optional[Any]:
        conversion_map: Dict[str, Callable[[str], Any]] = {
            "relation": partial(self._map_relation, col_key),
            "checkbox": map_checkbox,
            "date": map_notion_date,
            "created_time": map_date,
            "last_edited_time": map_date,
            "multi_select": split_str,
            "number": map_number,
            "file": self._map_file,
            "person": self._map_person,
        }

        try:
            return conversion_map[value_type](col_value)
        except KeyError:
            return col_value
        except TypeConversionError as e:
            if not col_value.strip():
                return None

            self._error(str(e))
            return None

    def _pop_column_type(self, row: CSVRowType, col_type_to_pop: str) -> Optional[Any]:
        """Some column types can't have multiple values (like created_time)
        so we pop them out of the row leaving only the last non-empty one"""

        last_col_value = None

        for col_key, col_value in list(row.items()):
            col_type = self.db.columns[col_key]["type"]

            if col_type == col_type_to_pop:
                result_value = self._map_column(col_key, col_value, col_type)

                self._raise_if_mandatory_empty(col_key, result_value)

                if result_value is not None:
                    last_col_value = result_value

                row.pop(col_key)

        return last_col_value

    def _map_icon(self, row: CSVRowType) -> Optional[FileType]:
        icon: Optional[FileType] = None

        if self.rules.icon_column:
            icon = row.get(self.rules.icon_column, "").strip()
            if icon:
                icon = map_icon(icon)
                if isinstance(icon, Path):
                    icon = self._relative_path(icon)

            self._raise_if_mandatory_empty(self.rules.icon_column, icon)

            if not self.rules.icon_column_keep:
                row.pop(self.rules.icon_column, None)

        if not icon and self.rules.default_icon:
            icon = self.rules.default_icon

        return icon

    def _map_image(self, row: CSVRowType) -> List:

            image: Optional[FileType] = None
            images = []

            if self.rules.image_column:
        
                for image_column in self.rules.image_column:
                    image = row.get(image_column, "").strip()
                    if image:
                        image = map_url_or_file(image)
                        
                        if isinstance(image, Path):
                            image = self._relative_path(image)

                    self._raise_if_mandatory_empty(image_column, image)

                    images.append(image)
                    if not self.rules.image_column_keep:
                        row.pop(image_column, None)
            
            return images
    
    def _map_non_cover_images(self,row:CSVRowType) ->  List:

        images = None

        if self.rules.image_column:
            image_columns = self._remove_cover_image(self.rules.image_caption_column)
            

        if self.rules.image_column:
            pass



    def _remove_cover_image(self,image_column:List) -> List:

        if len(image_column) == 1:
            return None
        
        img_col_copy = image_column.copy()
        img_col_copy.pop(0)

        return img_col_copy
    
    def _mention_cover_image(self,image_column:List) -> List:
	
        if len(image_column) == 1:
            return image_column
        
        img_col_copy = image_column.copy()
        cover_image = img_col_copy.pop(0)
        img_col_copy.append(cover_image)

        return img_col_copy

    def _map_image_caption(self, row: CSVRowType) -> Optional[str]:
        image_caption = None

        if self.rules.image_caption_column:
            image_caption = row.get(self.rules.image_caption_column, "").strip()

            self._raise_if_mandatory_empty(
                self.rules.image_caption_column, image_caption
            )

            if not self.rules.image_caption_column_keep:
                row.pop(self.rules.image_caption_column, None)

        return image_caption

    def _map_file(self, s: str) -> List[FileType]:

    
        col_value = split_str(s)

        resolved_uris = []
        for v in col_value:
            file_uri = map_url_or_file(v)

            if isinstance(file_uri, Path):
                rel_file_uri = self._ensure_path(file_uri)

                if rel_file_uri is None:
                    continue

                file_uri = rel_file_uri

            if file_uri not in resolved_uris:
                resolved_uris.append(file_uri)

        return resolved_uris

    def _ensure_path(self, path: Path) -> Optional[Path]:
        ensured_path = self._relative_path(path)
        if ensured_path is None:
            return None

        if _is_banned_extension(ensured_path):
            self._error(
                f"File extension '*{ensured_path.suffix}' is not allowed"
                f" to upload on csv2notion_neo.notion."
            )
            return None

        return ensured_path

    def _relative_path(self, path: Path) -> Optional[Path]:
        #PATH
        search_path = self.rules.files_search_path

        if not path.is_absolute():
            path = search_path / path

        if not path.exists():
            self._error(f"File {path.name} does not exist.")
            return None

        return path

    def _map_relation(
        self, relation_column: str, col_value: str
    ) -> List[CollectionRowBlockExtended]:
        col_values = split_str(col_value)

        resolved_relations = []
        for v in col_values:
            if v.startswith("https://www.csv2notion_neo.notion.so/"):
                resolved_relation = self._resolve_relation_by_url(relation_column, v)
            else:
                resolved_relation = self._resolve_relation_by_key(relation_column, v)

            if resolved_relation and resolved_relation not in resolved_relations:
                resolved_relations.append(resolved_relation)

        return resolved_relations

    def _resolve_relation_by_key(
        self, relation_column: str, key: str
    ) -> Optional[CollectionRowBlockExtended]:
        relation = self.db.relations[relation_column]

        try:
            return relation.rows[key]
        except KeyError:
            if self.rules.add_missing_relations:
                return relation.add_row_key(key)

            self._error(
                f"Value '{key}' for relation"
                f" '{relation_column} [column] -> {relation.name} [DB]'"
                f" is not a valid value."
            )

            return None

    def _resolve_relation_by_url(
        self, relation_column: str, url: str
    ) -> Optional[CollectionRowBlockExtended]:
        block_id = self._extract_id(url)
        if block_id is None:
            return None

        relation = self.db.relations[relation_column]
        relation_rows = relation.rows.values()

        try:
            return next(r for r in relation_rows if r.id == block_id)
        except StopIteration:
            self._error(
                f"Row with url '{url}' not found in relation"
                f" '{relation_column} [column] -> {relation.name} [DB]'."
            )

            return None

    def _extract_id(self, url: str) -> Optional[str]:
        try:
            return str(extract_id(url))
        except InvalidNotionIdentifier:
            self._error(f"'{url}' is not a valid csv2notion_neo.notion URL.")

            return None

    def _map_person(self, col_value: str) -> List[User]:
        col_values = split_str(col_value)

        resolved_persons = []
        for v in col_values:
            if is_email(v):
                resolved_person = self.db.users.get(v)

                if not resolved_person:
                    resolved_person = self.db.find_user(v)
            else:
                resolved_person = self.db.get_user_by_name(v)

            if resolved_person is None:
                self._error(f"Person '{v}' cannot be resolved.")
                continue

            if resolved_person not in resolved_persons:
                resolved_persons.append(resolved_person)

        return resolved_persons

    def _raise_if_mandatory_empty(self, col_key: str, col_value: Any) -> None:
        is_mandatory = col_key in self.rules.mandatory_column

        if is_mandatory and not col_value:
            raise NotionError(f"Mandatory column '{col_key}' is empty")


def _is_banned_extension(file_path: Path) -> bool:
    return file_path.suffix in {".exe", ".com", ".js"}
